/*
================================================================

        Alias-free digital synthesis of classic
        analog waveforms using BLIT-SWS method.

Implemented by Laurent de Soras, 1999
Based on "Alias-Free Digital Synthesis of Classic Analog
Waveforms" by Tim Stilson and Julius Smith, ICMC 1996.

Outputs a list of sample values stored into a text file which
can be read by CoolEdit using raw/text import. Not optimized,
intended to be only a proof of concept.

================================================================
Tab = 3
*/



#if defined (_MSC_VER)
#include	<crtdbg.h>
#endif

#include	<fstream>
#include	<string>

#include	<cassert>
#include	<climits>
#include	<cmath>



enum SignalType
{
	PULSE,
	SAWTOOTH,
	SQUARE,
	TRIANGLE		// Not very stable in the current implementation
};

enum BlitType
{
	STD_BLIT,	// Standard
	BP_BLIT		// Bipolar
};

struct Cycle
{
	// Construction
	Cycle ()
	:	_sign (1)
	,	_int (0)
	,	_frac (0)
	{}

	Cycle (double value, int sign)
	:	_sign (sign)
	,	_int (static_cast <long> (floor (value)))
	,	_frac (static_cast <unsigned long>
			((value - floor (value)) * (static_cast<double> (ULONG_MAX) + 1)))
	{}

	// Fixed point value
	long				_int;
	unsigned long	_frac;

	// Pulse sign
	int				_sign;
};



int	main (int argc, char *argv [])
{
#if ! defined (NDEBUG) && defined (_MSC_VER)
	{
		::_CrtSetReportMode (_CRT_WARN, _CRTDBG_MODE_WNDW);
		::_CrtSetReportMode (_CRT_ERROR, _CRTDBG_MODE_WNDW);
		::_CrtSetReportMode (_CRT_ASSERT, _CRTDBG_MODE_WNDW);

		const int	old_flags = ::_CrtSetDbgFlag (_CRTDBG_REPORT_FLAG);
		::_CrtSetDbgFlag (old_flags | _CRTDBG_LEAK_CHECK_DF);
	}
#endif

	using namespace std;

	// Signal parameters.
	const string	out_filename   = "out.txt";
	const SignalType	signal_type = TRIANGLE;		// Waveform
	const double	fs             = 44100.0;		// Sampling frequency
	const double	f0             =   220.0;		// Signal frequency
	const double	duty_cycle     = 1.0 / 3;		// Pulsewidth (for BP-BLIT based waveforms)
	const long		nbr_spl        = static_cast <long> (fs * 2.0);	// Nbr of samples to generate

	// Check parameters
	assert (fs > 0);
	assert (f0 > 0);
	assert (f0 <= fs / 2);
	assert (duty_cycle >= f0 / fs);
	assert (1 - duty_cycle >= f0 / fs);

	long				i;

/*______________________________________________
 *
 * Sinc look-up table
 *______________________________________________
 */

	// Look-up table parameters
	const int		SINC_RESOL = 4;	// Nbr of bits of the nbr of points/sample
	const long		SINC_ZEROS = 16;	// Nbr of zero-crossings.

	// Check parameters
	assert (SINC_RESOL > 0);
	assert (SINC_RESOL < sizeof (long) * CHAR_BIT - 2);
	assert (SINC_ZEROS > 0);
	assert (  static_cast <double> (SINC_ZEROS) * (1L << SINC_RESOL)
	        < static_cast <double> (LONG_MAX));

	const long		SINC_PTS_PER_SPL = 1L << SINC_RESOL;
	const long		SINC_NBR_SPL     = SINC_ZEROS * 2;
	const long		SINC_TABLE_LEN   = SINC_NBR_SPL * SINC_PTS_PER_SPL;
	const long		SINC_ZERO_OFFSET = SINC_TABLE_LEN >> 1;
	double			sinc_table [SINC_TABLE_LEN];	// Prefer allocation on the heap for big values

	const double	pi = atan (1) * 4;

	for (i = 0; i < SINC_TABLE_LEN; ++i)
	{
		if (i == SINC_ZERO_OFFSET)
		{
			sinc_table [i] = 1;
		}
		else
		{
			const double	pos      = (i - SINC_ZERO_OFFSET) * pi;
			const double	sinc_pos = pos * (1.0 / SINC_PTS_PER_SPL);
			const double	win_pos  = pos * (1.0 / SINC_TABLE_LEN);

			const double	sinc     = sin (sinc_pos) / sinc_pos;

			// Blackmann window
			const double	win      =   0.50 * cos (win_pos * 0.5)
								           + 0.08 * cos (win_pos)
								           + 0.42;

			sinc_table [i] = sinc * win;
		}
	}

/*______________________________________________
 *
 * Signal generation
 *______________________________________________
 */

	// Buffer init
	const long		BUFFER_LEN = SINC_NBR_SPL << 1;	// Must be a power of 2, and greater than SINC_NBR_SPL
	double			buffer [BUFFER_LEN];
	for (i = 0; i < BUFFER_LEN; ++i)
	{
		buffer [i] = 0;
	}

	BlitType			blit_type = STD_BLIT;
	if (   signal_type == SQUARE
	    || signal_type == TRIANGLE)
	{
		blit_type = BP_BLIT;
	}

	const int		POS_TO_PHASE = sizeof (long) * CHAR_BIT - SINC_RESOL;
	const double	dc_offset    = f0 / fs;		// For standard BLIT
	const double	period       = fs / f0;
	const double	integrator   = exp (log (0.8) / period);	// Leaky integrator

	Cycle				pos;
	Cycle				per [2];
	if (blit_type == BP_BLIT)
	{
		const double	c_len_0 = period * duty_cycle;
		const double	c_len_1 = period - c_len_0;

		per [0] = Cycle (c_len_0, +1);
		per [1] = Cycle (c_len_1, -1);
	}
	else
	{
		per [0] = Cycle (period, +1);
		per [1] = per [0];
	}

	// Initial values
	long				buf_pos   = 0;		// Current buffer positon
	int				cycle_cnt = 0;		// Current cycle (for BP-BLITs)
	double			square    = -duty_cycle;
	double			result    = -0.5;	// Initial signal value
	switch (signal_type)
	{
	case	PULSE:
		result = -dc_offset;
		break;
	case	SQUARE:
		result = -duty_cycle;
		break;
	};

	ofstream			out_file ("result.txt");

	for (long spl_cnt = -SINC_ZEROS; spl_cnt < nbr_spl; ++spl_cnt)
	{
		//	New pulse
		if (pos._int <= 0)
		{
			long		work_len = SINC_NBR_SPL;
			long		offset = SINC_PTS_PER_SPL - long (pos._frac >> POS_TO_PHASE);
			long		cur_pos = buf_pos;
			if (offset == SINC_PTS_PER_SPL)
			{
				offset = 0;
				-- cur_pos;
			}

			// Generate 1 pulse and mix it with the buffer
			do
			{
				assert (offset >= 0);
				assert (offset < SINC_TABLE_LEN);

				++ cur_pos;
				buffer [cur_pos & (BUFFER_LEN-1)] += sinc_table [offset] * pos._sign;
				offset += SINC_PTS_PER_SPL;
				-- work_len;
			}
			while (work_len > 0);

			// Next cycle (2 cycles = 1 period for BP-BLIT)
			pos._int   = per [cycle_cnt]._int;
			pos._frac += per [cycle_cnt]._frac;
			if (pos._frac < per [cycle_cnt]._frac)
			{
				++ pos._int;
			}
			cycle_cnt ^= 1;
			pos._sign  = per [cycle_cnt]._sign;
		}

		// Get current pulse value and clear buffer
		const double	pulse = buffer [buf_pos];
		buffer [buf_pos] = 0;

		// Skip first samples to avoid delay problems
		if (spl_cnt >= 0)
		{
			// Compute final signal
			switch (signal_type)
			{
			// BLIT based signals
			case	PULSE:
				result = pulse - dc_offset;
				break;
			case	SAWTOOTH:
				result = result * integrator + (pulse - dc_offset) * period / (period - 1);
				break;

			// BP-BLIT based signals
			case	SQUARE:
				result = result * integrator + pulse;
				break;
			case	TRIANGLE:
				square = square * integrator + pulse;
				result = result * integrator + square / (period * duty_cycle * (1 - duty_cycle));
				break;
			}

			// Store value into file (32-bit f.p data for CoolEdit)
			out_file << (result * (1L<<23)) << '\n';
		}

		// Next sample
		-- pos._int;
		buf_pos = (buf_pos + 1) & (BUFFER_LEN - 1);
	}

	out_file.close ();

	return (0);
}

